#!/bin/bash

IMXBOOT_DIR=$(dirname "$(realpath "$0")")
SECUREBOOT_DIR=$(realpath "$IMXBOOT_DIR/../secureboot_x2")
# keep SETUP_DIR for compat with older secureboot.conf (easier copy/paste)
SETUP_DIR="$IMXBOOT_DIR"

if [ -e "$SECUREBOOT_DIR/secureboot.conf" ]; then
	. "$SECUREBOOT_DIR/secureboot.conf"
elif [ -e "$IMXBOOT_DIR/secureboot.conf" ]; then
	. "$IMXBOOT_DIR/secureboot.conf"
else
	. "$IMXBOOT_DIR/secureboot.conf.example"
fi

# some defaults...
SWU_OUT_DIR=${SWU_OUT_DIR:-$SECUREBOOT_DIR/swu}
OUT=${OUT:-$SECUREBOOT_DIR/out}
CST=${CST:-$SECUREBOOT_DIR/cst}
IMXBOOT=${IMXBOOT:-$IMXBOOT_DIR}
# variables used in swu desc files need to be exported
# in particular SECUREBOOT_OUT is only used for mkswu
export TMPDIR=${TMPDIR:-$SECUREBOOT_DIR/tmp}
export SECUREBOOT_OUT="$OUT"

# constants
IMXBOOT_LOAD_ADDR=0x40480000
IMXBOOT_DEKBLOB_ADDR=0x40400000
LINUX_LOAD_ADDR=0x80080000
LINUX_FDT_ADDR=0x45000000
LINUX_FDT_OVERLAY_ADDR=0x45020000

CSF_HEADER="[Header]
Version = 4.3
Hash Algorithm = sha256
Engine = CAAM
Engine Configuration = 0
Certificate Format = X509
Signature Format = CMS

[Install SRK]
File = \"$CST/crts/SRK_1_2_3_4_table.bin\"
Source index = $CST_SOURCE_INDEX

[Install CSFK]
File = \"$CST/crts/CSF$((CST_SOURCE_INDEX+1))_1_sha256_${CST_KEYTYPE}_v3_usr_crt.pem\"

[Authenticate CSF]

[Install Key]
Verification index = 0
Target index = 2
File = \"$CST/crts/IMG$((CST_SOURCE_INDEX+1))_1_sha256_${CST_KEYTYPE}_v3_usr_crt.pem\""

case "$CST_UNLOCK_SRK" in
[1yY]*)
	CSF_HEADER="$CSF_HEADER

[Unlock]
Engine = OCOTP
Features = SRK REVOKE"
	;;
esac


LOG_REDIR=""

print_hint() {
	local rc="$?"
	if [ "$rc" != 0 ]; then
		to_stderr printf "%s\n" "" \
			"Command failed ($rc)." \
			"Please see $LOG_REDIR for details"
	fi
	return $rc
}

redirect_logs() {
	[ -n "$LOG_REDIR" ] && return

	trap print_hint EXIT

	# - prints to fd 3 will be displayed as stdout, fd 4 for stderr
	# - all other logs will go to log file instead
	LOG_REDIR="$TMPDIR/$cmd.log"
	echo "Logging build outputs to $LOG_REDIR"
	echo
	exec 4>&2 3>&1 >"$LOG_REDIR" 2>&1
}
to_stderr() {
	if [ -n "$LOG_REDIR" ]; then
		"$@" 2>&1 | tee /proc/$$/fd/4 >&2
	else
		"$@" >&2
	fi
}

to_stdout_only() {
	if [ -n "$LOG_REDIR" ]; then
		"$@" >&3 2>&1
	else
		"$@" 2>&1
	fi
}

to_stdout() {
	if [ -n "$LOG_REDIR" ]; then
		"$@" 2>&1 | tee /proc/$$/fd/3
	else
		"$@" 2>&1
	fi
}

error() {
	to_stderr printf "ERROR: %s\n" "$@"
	exit 1
}

warning() {
	[ "${debug:-2}" -lt 1 ] && return
	to_stderr printf "WARNING: %s\n" "$@"
}

info() {
	to_stdout printf "%s\n" "$@"
}

ERRORS=""
soft_error() {
	warning "$@"
	ERRORS="$ERRORS
$*"
	return 1
}

prompt_yesno() {
	local default="$1"
	shift 1
	local yesno

	while true; do
		to_stdout_only read -r -p "$1" yesno
		[ -n "$yesno" ] || yesno="$default"
		case "$yesno" in
			[Yy] | [Yy][Ee][Ss] | [Tt][Rr][Uu][Ee] | 1)
				return 0
				;;
			[Nn] | [Nn][Oo] | [Ff][Aa][Ll][Ss][Ee] | 0)
				return 1
				;;
		esac
	done
}

check_cst_srk() {
        [[ -e "$CST/crts/SRK_1_2_3_4_table.bin" ]] \
                || error "SRK not setup yet, please run secureboot.sh init_srk first"
        [[ -e "$CST/keys/SRK$((CST_SOURCE_INDEX+1))_sha256_${CST_KEYTYPE}_v3_ca_key.pem" ]] \
                || error "missing SRK key $CST/keys/SRK$((CST_SOURCE_INDEX+1))_sha256_${CST_KEYTYPE}_v3_ca_key.pem"
        [[ -e "$CST/keys/IMG1_1_sha256_${CST_KEYTYPE}_v3_usr_key.pem" ]] \
                || error "SRK was not init for habv4 (Armadillo IoT G4/X2's imx8mp), extract directory anew and rerun secureboot.sh init_srk"
}

pre_checks() {
	local overlay cmd="$1"
	[ -n "$OUT" ] || error "\$OUT is not set"
	[ -n "$CST" ] || error "\$CST is not set"
	[ -n "$SWU_OUT_DIR" ] || error "\$SWU_OUT_DIR is not set"
	[ -n "$TMPDIR" ] || error "\$TMPDIR is not set"

	mkdir -p "$OUT" "$CST" "$SWU_OUT_DIR" "$TMPDIR" || exit

	# Special commands that need to skip checks
	case "$cmd" in
	setup|print_srk_fuse) return;;
	esac

	if [ -e "$CST/linux64/bin/cst" ]; then
		CST_BIN=$CST/linux64/bin/cst
	elif command -v cst >/dev/null; then
		CST_BIN=cst
	else
		error "Please run $0 setup"
	fi

	case "$cmd" in
	secureboot_init|make_installer)
		# check setup environment
		[[ -e "$SECUREBOOT_DIR/build-rootfs" ]] \
			|| soft_error "could not find build-rootfs"
		[[ -e "$SECUREBOOT_DIR/imx-boot" ]] \
			|| soft_error "could not find imx_boot"

		command -v mkswu > /dev/null || soft_error "could not find mkswu command"

		[[ -z "$ERRORS" ]] || error "Please run $0 setup"

		if [[ "$(realpath "$SECUREBOOT_DIR/secureboot.sh")" != "$(realpath "$IMXBOOT_DIR/secureboot.sh")" ]]; then
			info "This script is not $(readlink "$SECUREBOOT_DIR/secureboot.sh") that was used during setup."
			prompt_yesno y "Update ? [Y/n]" || error "Aborted by user"
			ln -sf "$IMXBOOT_DIR/secureboot.sh" "$SECUREBOOT_DIR/secureboot.sh"
			ln -sf "$IMXBOOT_DIR" "$SECUREBOOT_DIR/imx-boot"
		fi
		[[ -e "$IMXBOOT" ]] || error "\$IMXBOOT does not exist: $IMXBOOT. Remove variable assignment?"
		check_or_download_linux
		;;
	imxboot*)
		check_cst_srk
		[[ -e "$IMXBOOT" ]] || error "\$IMXBOOT does not exist: $IMXBOOT. Remove variable assignment?"
		;;
	linux*)
		check_cst_srk
		check_or_download_linux
		;;
	esac

	redirect_logs
}

init_srk() {
	# already done
	if [[ -e "$CST/crts/SRK_1_2_3_4_table.bin" ]]; then
		info "Secure boot signing keys already setup"
		return
	fi

	info "Initializing secure boot signing keys..."

	mkdir -p "$CST/keys" "$CST/crts" || exit
	if ! [ -e "$CST/ca" ]; then
		cp -a "$IMXBOOT_DIR/secureboot/cst_ca" "$CST/ca" || exit
	fi

	(
		cd "$CST/keys" || exit
		PKI_SCRIPT=hab4_pki_tree.sh
		if ! [ -e "$PKI_SCRIPT" ]; then
			PKI_SCRIPT=/usr/share/doc/imx-code-signing-tool/pki_scripts/hab4_pki_tree.sh
		fi
		echo 00 > serial || error "Could not write serial"
		if ! [[ -e "key_pass.txt" ]]; then
			info "Generating random passphrase for CST keys (written as plain text in $CST/keys/key_pass.txt)"
			KEYPASS=$(tr -cd '0-9a-zA-Z#@!.[]_-' < /dev/urandom  | head -c 20)
			printf "%s\n" "$KEYPASS" "$KEYPASS" > key_pass.txt \
				|| error "Could not write key_pass.txt"
		fi
		# new 3.3.2 syntax
		ARGS=( -existing-ca n -num-srk 4 -srk-ca y )
		if grep -q -- '-kt' "$PKI_SCRIPT"; then
			case "$CST_ECC" in
			y) ARGS+=( -kt ecc );;
			*) ARGS+=( -kt rsa );;
			esac
		else
			ARGS+=( -use-ecc "$CST_ECC" )
		fi
		ARGS+=( -kl "$CST_KEYLEN" -duration "$CST_DURATION" )
		sh "$PKI_SCRIPT" "${ARGS[@]}"
	) || error "Could not generate hab4 pki tree"
	(
		cd "$CST/crts" || exit
		SRKTOOL=../linux64/bin/srktool
		if ! [ -e "$SRKTOOL" ]; then
			SRKTOOL=srktool
		fi
		"$SRKTOOL" --hab_ver 4 --digest sha256 \
		        --table SRK_1_2_3_4_table.bin --efuses SRK_1_2_3_4_fuse.bin \
			--fuse_format 1 \
			--certs "./SRK1_sha256_${CST_KEYTYPE}_v3_ca_crt.pem,./SRK2_sha256_${CST_KEYTYPE}_v3_ca_crt.pem,./SRK3_sha256_${CST_KEYTYPE}_v3_ca_crt.pem,./SRK4_sha256_${CST_KEYTYPE}_v3_ca_crt.pem"
	) || error "Could not generate SRK table/fuse map"

	echo
	info "SRK has been initialized. Backup $CST to a secure location !!"
}

print_srk_fuse() {
	local -a FUSES

	[ -e "$CST/crts/SRK_1_2_3_4_fuse.bin" ] \
		|| error "Secure boot key not initialized! Run $0 init_srk first"

	mapfile -t FUSES < <(hexdump -e '1/4 "0x"' -e '1/4 "%X""\n"' "$CST/crts/SRK_1_2_3_4_fuse.bin")
	local IFS=" "
	echo
	echo "Use the following commands in uboot to set root key hash in OTP:"
	echo "fuse prog -y 6 0 ${FUSES[*]:0:4}"
	echo "fuse prog -y 7 0 ${FUSES[*]:4:4}"
	echo "fuse prog -y 0 0 0x200"
	echo
	echo
	echo "For testing it is possible to temporarly override fuses instead,"
	echo "but this only validates the last part of the boot:"
	echo "fuse override 6 0 ${FUSES[*]:0:4}"
	echo "fuse override 7 0 ${FUSES[*]:4:4}"
}

# imxboot helper: get offset from build log
get_offset() {
	local key="$1"
	local target="$2"

	awk "/OFFSET dump/ { dump=1 }
	     dump && /make/ { dump=0 }
	     dump && /$key/ { print $target }" "$TMPDIR/$IMXBOOT_TARGET.buildlog"
}

#########################################################################
# normal secure boot, imxboot image                                     #
#                                                                       #
# build imxboot image and sign it                                       #
#########################################################################

imxboot() {
	info "Building imx-boot (boot loader)..."

	to_stdout_only make -s firmware  -C "$IMXBOOT" \
		|| error "Could not get firmwares"
	make -j5 -C "$IMXBOOT" "imx-boot_$IMXBOOT_TARGET" 2>&1 \
			| tee "$TMPDIR/$IMXBOOT_TARGET.buildlog" \
		|| error "Build of imxboot failed"

	local block
	# shellcheck disable=SC2016 ## single quotes are on purpose
	block=$(get_offset "spl hab block" '$4, $5, $6')
	[[ -n "$block" ]] || error "Could not get spl hab block from build logs"
	cat > "$TMPDIR/$IMXBOOT_TARGET.csf_spl.txt" <<EOF \
		|| error "Could not write $TMPDIR/$IMXBOOT_TARGET.csf_spl.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks =  $block "$IMXBOOT/imx-boot_$IMXBOOT_TARGET"

[Unlock]
Engine = CAAM
Features = MID
EOF
	{
		cat <<EOF
$CSF_HEADER

[Authenticate Data]
Verification index = 2
EOF
		# shellcheck disable=SC2016 ## single quotes are on purpose
		block=$(get_offset "sld hab block" '$4, $5, $6')
		[[ -n "$block" ]] || error "Could not get sld hab block from build logs"
		echo -n "Blocks = $block \"$IMXBOOT/imx-boot_$IMXBOOT_TARGET\"";
		while read -r block; do
			echo ", \\"
			echo -n "$block \"$IMXBOOT/imx-boot_$IMXBOOT_TARGET\""
		done < <(awk '/\.\/print_fit_hab\.sh/ { dump=1; getline }
			      dump && /make/ { dump=0 }
			      dump' "$TMPDIR/$IMXBOOT_TARGET.buildlog")
		echo
	} > "$TMPDIR/$IMXBOOT_TARGET.csf_fit.txt" \
		|| error "Could not write $TMPDIR/$IMXBOOT_TARGET.csf_fit.txt"

	"$CST_BIN" -i "$TMPDIR/$IMXBOOT_TARGET.csf_spl.txt" -o "$TMPDIR/$IMXBOOT_TARGET.csf_spl.bin" \
		|| error "Could not sign csf spl"
	"$CST_BIN" -i "$TMPDIR/$IMXBOOT_TARGET.csf_fit.txt" -o "$TMPDIR/$IMXBOOT_TARGET.csf_fit.bin" \
		|| error "Could not sign csf fit"
	cp "$IMXBOOT/imx-boot_$IMXBOOT_TARGET" "$TMPDIR/imx-boot_$IMXBOOT_TARGET.signed"
	local offset
	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ *csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get csf_off from build logs"
	dd if="$TMPDIR/$IMXBOOT_TARGET.csf_spl.bin" of="$TMPDIR/imx-boot_$IMXBOOT_TARGET.signed" \
		seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc

	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ sld_csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get sld_csf_off from build logs"
	dd if="$TMPDIR/$IMXBOOT_TARGET.csf_fit.bin" of="$TMPDIR/imx-boot_$IMXBOOT_TARGET.signed" \
		seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc

	mv "$TMPDIR/imx-boot_$IMXBOOT_TARGET.signed" \
		"$OUT/imx-boot_$IMXBOOT_TARGET.signed" || exit
	info "Created $OUT/imx-boot_$IMXBOOT_TARGET.signed"
}

#########################################################################
# normal secure boot, linux image                                       #
#                                                                       #
# build linux fit image from configured settings and sign it            #
#########################################################################

linux_apply_overlays() {
	if [[ "${#LINUX_DTB_OVERLAYS_PREAPPLY[@]}" = 0 ]]; then
		cp "$LINUX_DTB" "$TMPDIR/$dtb" || error "Could not copy dtb"
		return
	fi

	# add prefix
	local overlay
	local -a overlays
	for overlay in "${LINUX_DTB_OVERLAYS_PREAPPLY[@]}"; do
		if [[ "${overlay#/}" != "$overlay" ]]; then
			overlays+=( "$overlay" )
		else
			overlays+=( "${LINUX_DTB%/*}/$overlay" )
		fi
	done

	fdtoverlay -v -i "$LINUX_DTB" "${overlays[@]}" \
			-o "$TMPDIR/$dtb" \
		|| error "Could not apply overlays"
}

maybe_compress() {
	local file
	for file; do
		[[ -e "$file" ]] || continue
		[[ -s "$file.zst" ]] && [[ "$file.zst" -nt "$file" ]] && continue
		zstd -c "$file" > "$file.zst" \
			|| error "Could not compress $file"
	done
}

# helper: build fit image from linux image and dtbs
linux_build_fit() {
	local overlay

	MKIMAGE="$IMXBOOT/uboot-imx/$IMXBOOT_TARGET/tools/mkimage"
	if ! [ -e "$MKIMAGE" ]; then
		if command -v mkimage >/dev/null; then
			MKIMAGE=mkimage
		else
			error "mkimage not found, build imxboot first"
		fi
	fi

	# Note: each entry could have a 'hash-1 { algo = "sha1"; };' block, but
	# since we sign content it is redundant.
	# The hash might also not be compatible with encryption as we might
	# only encrypt the content and verification would happen too early
	{
		cat <<EOF
/dts-v1/;
/ {
	description = "fit linux image for armadillo";
	#address-cells = <1>;
	images {
		kernel {
			description = "linux kernel";
			data = /incbin/("$LINUX_IMAGE.zst");
			type = "kernel";
			arch = "arm64";
			os = "linux";
			compression = "zstd";
			load = <$LINUX_LOAD_ADDR>;
			entry = <$LINUX_LOAD_ADDR>;
		};
		fdt {
			description = "fdt";
			data = /incbin/("$TMPDIR/$dtb.zst");
			type = "flat_dt";
			arch = "arm64";
			compression = "zstd";
			load = <$LINUX_FDT_ADDR>;
		};
EOF
		if [[ -n "$LINUX_INITRD" ]]; then
			cat <<EOF
		ramdisk {
			description = "initrd";
			data = /incbin/("$LINUX_INITRD");
			type = "ramdisk";
			arch = "arm64";
			os = "linux";
			compression = "none";
		};
EOF
		fi
		for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
			fullpath="${LINUX_DTB%/*}/$overlay"
			cat <<EOF
		$overlay {
			description = "$overlay";
			data = /incbin/("$fullpath.zst");
			type = "flat_dt";
			arch = "arm64";
			compression = "zstd";
			load = <$LINUX_FDT_OVERLAY_ADDR>;
		};
EOF
		done
		cat <<EOF
	};
	configurations {
		default = "armadillo";
		armadillo {
			description = "armadillo";
			kernel = "kernel";
			fdt = "fdt";
EOF
		if [[ -n "$LINUX_INITRD" ]]; then
			cat <<EOF
			ramdisk = "ramdisk";
EOF
		fi
			cat <<EOF
		};
EOF
		for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
			cat <<EOF
		$overlay {
			fdt = "$overlay";
		};
EOF
		done
		cat <<EOF
	};
};
EOF
	} > "$TMPDIR/Image.its" \
		|| error "Could not write linux fit configuration"

	maybe_compress "$LINUX_IMAGE" "$TMPDIR/$dtb"
	for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
		maybe_compress "${LINUX_DTB%/*}/$overlay"
	done

	"$MKIMAGE" -f "$TMPDIR/Image.its" "$TMPDIR/Image.fit" \
		|| error "Could not build linux fit image"
}

build_initrd() {
	[ -e "$LINUX_INITRD" ] && return

	while [ -z "$LINUX_INITRD_TYPE" ]; do
		info "Select which initrd to build: [none|plain|mmc|verity]"
		to_stdout_only read -r LINUX_INITRD_TYPE
		case "$LINUX_INITRD_TYPE" in
		none|plain|mmc) ;;
		verity)
			[ -e "$LINUX_INITRD_VERITY_CERT" ] \
				|| error "verity initrd requires LINUX_INITRD_VERITY_CERT to be set";;
		*) LINUX_INITRD_TYPE="";;
		esac
	done

	local build_opt=()

	case "$LINUX_INITRD_TYPE" in
	none)
		LINUX_INITRD=""
		return
		;;
	plain)
		;;
	mmc)
		build_opt=( "--lock" )
		;;
	verity)
		build_opt=( --verity "$LINUX_INITRD_VERITY_CERT" )
		;;
	*)
		error "invalid LINUX_INITRD_TYPE $LINUX_INITRD_TYPE"
		;;
	esac

	LINUX_INITRD="$TMPDIR/initrd-baseos.$LINUX_INITRD_TYPE.zst"
	info "Building initrd for $LINUX_INITRD_TYPE (first time is slow)"
	(
		cd "$SECUREBOOT_DIR/build-rootfs" || exit
		./build_initrd.sh "$LINUX_INITRD" "${build_opt[@]}"|| exit
		[ -e "$LINUX_INITRD" ]
	) || error "could not build initrd"
}

linux() {
	local dtb=${LINUX_DTB##*/}
	local linux_size
	local OUTPUT OUTPUT_TYPE=""

	build_initrd
	case "${LINUX_INITRD##*/}" in
	*".mmc."*)
		OUTPUT_TYPE="-mmc"
		;;
	*".verity."*)
		OUTPUT_TYPE="-verity"
		;;
	esac
	OUTPUT="Image.signed$OUTPUT_TYPE"

	info "Signing linux image${OUTPUT_TYPE:+ (${OUTPUT_TYPE#-})}..."

	linux_apply_overlays
	linux_build_fit
	linux_size=$(stat -c "%s" "$TMPDIR/Image.fit")
	[[ -n "$linux_size" ]] || error "Could not get linux image size"
	# round up to 0x1000 as pert uboot authenticate_image()
	linux_size=$((((linux_size + 0xfff) / 0x1000) * 0x1000))
	mv "$TMPDIR/Image.fit" "$TMPDIR/Image_signing.bin" \
		|| error "Could not copy FIT image"
	cat > "$TMPDIR/genIVT.pl" <<EOF
#!/usr/bin/perl -w
use strict;
print pack("V", 0x412000D1); # Signature
print pack("V", $IMXBOOT_LOAD_ADDR); # Load Address (*load_address)
print pack("V", 0x0); # Reserved
print pack("V", 0x0); # DCD pointer
print pack("V", 0x0); # Boot Data
print pack("V", $(printf "%#x" $((IMXBOOT_LOAD_ADDR + linux_size)))); # Self Pointer (*ivt)
print pack("V", $(printf "%#x" $((IMXBOOT_LOAD_ADDR + linux_size + 0x20)))); # CSF Pointer (*csf)
print pack("V", 0x0); # Reserved
EOF
	perl "$TMPDIR/genIVT.pl" \
			| dd of="$TMPDIR/Image_signing.bin" bs=1k status=none \
				seek=$linux_size oflag=seek_bytes \
		|| error "Could not generate linux IVT"

	cat > "$TMPDIR/Image.csf_linux.txt" <<EOF \
		|| error "Could not write $TMPDIR/Image.csf_linux.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks = $IMXBOOT_LOAD_ADDR 0x00000000 $(printf "%#x" $((linux_size + 0x20))) "$TMPDIR/Image_signing.bin"
EOF
	"$CST_BIN" -i "$TMPDIR/Image.csf_linux.txt" -o "$TMPDIR/Image.csf_linux.bin" \
		|| error "Could not sign csf linux"
	cat "$TMPDIR/Image.csf_linux.bin" >> "$TMPDIR/Image_signing.bin" \
		|| error "Could not append to $TMPDIR/Image_signing.bin"
	chmod 644 "$TMPDIR/Image_signing.bin" \
		|| error "Could not fix image permission"
	mv "$TMPDIR/Image_signing.bin" "$OUT/$OUTPUT" \
		|| error "Could not rename to $OUT/$OUTPUT"
	info "Created $OUT/$OUTPUT"
}


#########################################################################
# encrypted secure boot, imxboot image                                  #
#                                                                       #
# build imxboot image, encrypt it in multiple steps and sign the result #
# 'dek key' is included as is in encrypted image and must be processed  #
# on each device before use                                             #
#########################################################################

imxboot_enc() {
	# encryption is known to NOT work with many cst versions, require 3.3.1
	if ! "$CST_BIN" --version | grep -q 3.3.1; then
		error "imxboot encryption is only supported with cst 3.3.1, contact support if really required"
	fi

	## rebuild imx-boot image to get build logs
	# for encryption we need dek_blob_fit_dummy.bin to be present to reserve
	# space in fit image for final dek blob
	mkdir -p "$IMXBOOT/imx-mkimage/$IMXBOOT_TARGET"
	dd if=/dev/zero of="$IMXBOOT/imx-mkimage/$IMXBOOT_TARGET/dek_blob_fit_dummy.bin" \
			bs=96 count=1 status=none \
		|| error "Could not create dummy dek blob for fit image"
	make -j5 -C "$IMXBOOT" "imx-boot_$IMXBOOT_TARGET" 2>&1 \
			| tee "$TMPDIR/$IMXBOOT_TARGET.buildlog" \
		|| error "Build of imxboot failed"
	# remove dummy dek blob as it would affect normal builds that could follow
	rm -f "$IMXBOOT/imx-mkimage/$IMXBOOT_TARGET/dek_blob_fit_dummy.bin"


	## generate csf configuration files for SPL uboot
	# note we cannot both sign and encrypt in a single step so we need
	# to generate two csf files
	local block
	# shellcheck disable=SC2016 ## single quotes are on purpose
	block=$(get_offset "spl hab block" '$4, $5, $6')
	[[ -n "$block" ]] || error "Could not get spl hab block from build logs"
	local block_start=${block%% *}
	local block_length=${block##* }
	local blob_address
	# Authenticate Start Address +  SPL & DDR FW image length + CSF Padding
	blob_address=$(printf "%#x" $((block_start + block_length + 0x2000)))
	local decrypt_block
	# Start Address = Start Address + SPL header = 0x91ffc0 + 0x40 = 0x920000
	# Offset = Image offset (image_off) = 0x40
	# Decrypt size = Image length - SPL header = 0x24a00 - 0x40 = 0x249c0
	decrypt_block=$(printf "%#x 0x40 %#x" \
			$((block_start + 0x40)) \
			$((block_length - 0x40)))
	cat > "$TMPDIR/$IMXBOOT_TARGET.csf_spl_enc.txt" <<EOF \
		|| error "Could not write $TMPDIR/$IMXBOOT_TARGET.csf_spl_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks =  $block_start 0x0 0x40 "$IMXBOOT/imx-boot_$IMXBOOT_TARGET"

[Unlock]
Engine = CAAM
Features = MID

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$TMPDIR/$IMXBOOT_TARGET.dek_spl.bin"
    Key Length = 256
    # Authenticate Start Address +  SPL & DDR FW image length + CSF Padding
    Blob Address = $blob_address

[Decrypt Data]
    # The decrypt data command below causes CST to modify the input
    # file and encrypt the specified block of data. This image file
    # is a copy of the file used for the authentication command above
    Verification Index = 0
    Mac Bytes = 16
    # Start Address = Start Address + SPL header
    # Offset = Image offset (image_off) = 0x40
    # Decrypt size = Image length - SPL header
    Blocks = $decrypt_block "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc"
EOF
	cat > "$TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.txt" <<EOF \
		|| error "Could not write $TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks =  $block "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc"

[Unlock]
Engine = CAAM
Features = MID

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$TMPDIR/$IMXBOOT_TARGET.dek_spl_dummy.bin"
    Key Length = 256
    # Authenticate Start Address +  SPL & DDR FW image length + CSF Padding
    Blob Address = $blob_address

[Decrypt Data]
    # The decrypt data command below causes CST to modify the input
    # file and encrypt the specified block of data. This image file
    # is a copy of the file used for the authentication command above
    Verification Index = 0
    Mac Bytes = 16
    # Start Address = Start Address + SPL header
    # Offset = Image offset (image_off) = 0x40
    # Decrypt size = Image length - SPL header
    Blocks = $decrypt_block "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy"
EOF


	## generate csf configuration file for FIT image (2nd stage:
	# uboot, optee, atf)

	# shellcheck disable=SC2016 ## single quotes are on purpose
	block=$(get_offset "sld hab block" '$4, $5, $6')
	[[ -n "$block" ]] || error "Could not get sld hab block from build logs"
	local fit_line1 fit_line2 fit_line3 fit_line4
	{
		read -r fit_line1
		read -r fit_line2
		read -r fit_line3
		read -r fit_line4
	} < <(awk '/\.\/print_fit_hab\.sh/ { dump=1; getline }
		   dump && /make/ { dump=0 }
		   dump' "$TMPDIR/$IMXBOOT_TARGET.buildlog")
	[[ -n "$fit_line1" ]] && [[ -n "$fit_line2" ]] \
		&& [[ -n "$fit_line3" ]] && [[ -n "$fit_line4" ]] \
		|| error "print_fit_hab format not as expected"
	# blocks need to be 16bit aligned, but first two lines are not
	# (and happen to be contiguous): merge these. Geez.
	local fit_line1_start fit_line1_offset fit_line1_len
	fit_line1_start=${fit_line1%% *}
	fit_line1_offset=${fit_line1% *}
	fit_line1_offset=${fit_line1_offset##* }
	fit_line1_len=${fit_line1##* }
	local fit_line2_start fit_line2_offset fit_line2_len
	fit_line2_start=${fit_line2%% *}
	fit_line2_offset=${fit_line2% *}
	fit_line2_offset=${fit_line2_offset##* }
	fit_line2_len=${fit_line2##* }
	# check 2nd line start/offset are 1st line + length
	(( fit_line2_start == fit_line1_start + fit_line1_len )) \
		&& (( fit_line2_offset == fit_line1_offset + fit_line1_len )) \
		|| error "print_fit_hab first two lines are not contiguous"
	fit_line1=$(printf "%s %s %#x" "$fit_line1_start" "$fit_line1_offset" \
			"$((fit_line1_len + fit_line2_len))")

	cat > "$TMPDIR/$IMXBOOT_TARGET.csf_fit_enc.txt" <<EOF \
		|| error "Could not write $TMPDIR/$IMXBOOT_TARGET.csf_fit_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks = $block "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc"

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$TMPDIR/$IMXBOOT_TARGET.dek_fit.bin"
    Key Length = 256
    Blob Address = $IMXBOOT_DEKBLOB_ADDR

[Decrypt Data]
    Verification Index = 0
    Mac Bytes = 16
    Blocks = $fit_line1 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc", \\
             $fit_line3 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc", \\
             $fit_line4 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc"
EOF
	cat > "$TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.txt" <<EOF \
		|| error "Could not write $TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks = $block "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc", \\
         $fit_line1 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc", \\
         $fit_line3 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc", \\
         $fit_line4 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc"

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$TMPDIR/$IMXBOOT_TARGET.dek_fit_dummy.bin"
    Key Length = 256
    Blob Address = $IMXBOOT_DEKBLOB_ADDR

[Decrypt Data]
    Verification Index = 0
    Mac Bytes = 16
    Blocks = $fit_line1 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy", \\
             $fit_line3 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy", \\
             $fit_line4 "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy"
EOF


	## encrypt then sign the SPL image

	cp "$IMXBOOT/imx-boot_$IMXBOOT_TARGET" "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" \
		|| error "Could not copy imx-boot_$IMXBOOT_TARGET"

	"$CST_BIN" -i "$TMPDIR/$IMXBOOT_TARGET.csf_spl_enc.txt" \
			-o "$TMPDIR/$IMXBOOT_TARGET.csf_spl_enc.bin" \
		|| error "Could not encrypt csf spl"

	cp "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy" \
		|| error "Could not copy imx-boot_$IMXBOOT_TARGET.enc.dummy"
	"$CST_BIN" -i "$TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.txt" \
			-o "$TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.bin" \
		|| error "Could not sign encrypted csf spl"
	rm -f "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy" "$TMPDIR/$IMXBOOT_TARGET.dek_spl_dummy.bin"


	## copy encryption nounce from first csf output into final signed one

	local offset_in offset_out
	offset_in=$(stat -c "%s" "$TMPDIR/$IMXBOOT_TARGET.csf_spl_enc.bin")
	[[ -n "$offset_in" ]] || error "Could not get csf_spl_enc.bin size"
	offset_in=$((offset_in - 36))
	offset_out=$(stat -c "%s" "$TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.bin")
	[[ -n "$offset_out" ]] || error "Could not get csf_spl_sign_enc.bin size"
	offset_out=$((offset_out - 36))
	echo "copying csf_spl_enc.bin $offset_in -> csf_spl_sign_enc.bin $offset_out 36 bytes"
	dd if="$TMPDIR/$IMXBOOT_TARGET.csf_spl_enc.bin" \
			of="$TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.bin" \
			status=none bs=36 count=1 \
			iflag=skip_bytes oflag=seek_bytes \
			skip="$offset_in" seek="$offset_out" \
		|| error "Could not copy nonce/mac to csf_spl_sign_enc"

	## encrypt then sign the FIT image

	"$CST_BIN" -i "$TMPDIR/$IMXBOOT_TARGET.csf_fit_enc.txt" \
			-o "$TMPDIR/$IMXBOOT_TARGET.csf_fit_enc.bin" \
		|| error "Could not encrypt csf fit"
	cp "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy" \
		|| error "Could not copy imx-boot_$IMXBOOT_TARGET.enc.dummy"
	"$CST_BIN" -i "$TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.txt" \
			-o "$TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.bin" \
		|| error "Could not sign encrypted csf fit"
	rm -f "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc.dummy" "$TMPDIR/$IMXBOOT_TARGET.dek_fit_dummy.bin"

	## copy encryption nounce from first csf output into final signed one

	offset_in=$(stat -c "%s" "$TMPDIR/$IMXBOOT_TARGET.csf_fit_enc.bin")
	[[ -n "$offset_in" ]] || error "Could not get csf_fit_enc.bin size"
	offset_in=$((offset_in - 36))
	offset_out=$(stat -c "%s" "$TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.bin")
	[[ -n "$offset_out" ]] || error "Could not get csf_fit_sign_enc.bin size"
	offset_out=$((offset_out - 36))
	echo "copying csf_fit_enc.bin $offset_in -> csf_fit_sign_enc.bin $offset_out 36 bytes"
	dd if="$TMPDIR/$IMXBOOT_TARGET.csf_fit_enc.bin" \
			of="$TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.bin" \
			status=none bs=36 count=1 \
			iflag=skip_bytes oflag=seek_bytes \
			skip="$offset_in" seek="$offset_out" \
		|| error "Could not copy nonce/mac to csf_fit_sign_enc"


	# write final spl csf output into boot image

	local offset
	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get csf_off from build logs"
	: "copying csf_spl_sign_enc.bin (full) -> final boot image $offset"
	dd if="$TMPDIR/$IMXBOOT_TARGET.csf_spl_sign_enc.bin" \
		of="$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" \
		status=none seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc


	## write spl dek raw key into boot image
	# note we write dek_spl here but it is not final:
	# it needs to be encapsulated for the device on the device
	# we nevertheless use the same emplacement in the file for convenience
	# as encapsulated version is bigger and will overwrite this fully
	# dek offsets are stored in dek_offsets file for swupdate/manual flashing


	offset=$((offset + 0x2000))
	printf "offset for dek_spl_blob: %#x\n" "$offset"
	printf "dek_spl_offset %#x " "$offset" > "$TMPDIR/$IMXBOOT_TARGET.dek_offsets"
	dd if="$TMPDIR/$IMXBOOT_TARGET.dek_spl.bin" \
		of="$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" \
		status=none seek="$offset" oflag=seek_bytes bs=4K conv=notrunc


	## same with fit csf binary and dek

	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ sld_csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get sld_csf_off from build logs"
	: "copying csf_fit_sign_enc.bin (full) -> final boot image $offset"
	dd if="$TMPDIR/$IMXBOOT_TARGET.csf_fit_sign_enc.bin" \
		of="$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" \
		status=none seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc

	offset=${fit_line4#* }
	offset=$((${offset#* } + ${offset% *}))
	printf "offset for dek_fit_blob: %#x\n" "$offset"
	printf "dek_fit_offset %#x\n" "$offset" >> "$TMPDIR/$IMXBOOT_TARGET.dek_offsets"
	dd if="$TMPDIR/$IMXBOOT_TARGET.dek_fit.bin" \
		of="$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" \
		status=none seek="$offset" oflag=seek_bytes bs=4K conv=notrunc

	mv "$TMPDIR/imx-boot_$IMXBOOT_TARGET.enc" \
		"$TMPDIR/$IMXBOOT_TARGET.dek_offsets" \
		"$OUT" || exit

	info "Created $OUT/imx-boot_$IMXBOOT_TARGET.enc"
}

setup() {
	local build_rootfs_dir=""

	if ! [ -e "$IMXBOOT_DIR/.git" ]; then
		git -C "$IMXBOOT_DIR" init -b master -q
		git -C "$IMXBOOT_DIR" add .
		git -C "$IMXBOOT_DIR" commit -q -m "init from secureboot.sh setup"
	fi

	ln -sf "$IMXBOOT_DIR/secureboot.sh" "$SECUREBOOT_DIR/secureboot.sh" \
		|| error "symbolic link failed: secureboot.sh"

	if [ ! -e "$SECUREBOOT_DIR/secureboot.conf" ]; then
		cp "$IMXBOOT_DIR/secureboot.conf.example" "$SECUREBOOT_DIR/secureboot.conf" \
			|| error "could not create $SECUREBOOT_DIR/secureboot.conf"
	fi

	# no cst package
	if ! [ -e "$CST/linux64/bin/cst" ] && ! command -v cst > /dev/null; then
		info "imx-code-signing-tool package is needed to setup secureboot."
		prompt_yesno y "Install imx-code-signing-tool? [Y/n]" \
			|| error "Aborted by user"
		sudo apt install -y imx-code-signing-tool
	fi

	local build_rootfs_setup_dir="$SECUREBOOT_DIR/build-rootfs"
	if [ ! -e "$build_rootfs_setup_dir" ]; then
		build_rootfs_dir="$(find .. -maxdepth 1 -type d -name "build-rootfs-*" | sort -Vr | head -n 1)"
		# download latest if none found
		if [ -z "$build_rootfs_dir" ]; then
			info "Downloading latest build-rootfs..."
			curl -O https://download.atmark-techno.com/armadillo-iot-g4//tool/build-rootfs-latest.tar.gz \
				|| soft_error "failed to download: https://download.atmark-techno.com/armadillo-iot-g4//tool/build-rootfs-latest.tar.gz" \
			&& tar -C "$SECUREBOOT_DIR" -xaf build-rootfs-latest.tar.gz \
				|| soft_error "tar failed: build-rootfs-latest.tar.gz" \
			&& build_rootfs_dir="$SECUREBOOT_DIR/$(tar tf build-rootfs-latest.tar.gz \
					| awk -F '/' '{ print $1; exit; }')"
			git -C "$build_rootfs_dir" init -b master -q
			git -C "$build_rootfs_dir" add .
			git -C "$build_rootfs_dir" commit -q -m "init from secureboot.sh setup"
			rm -rf build-rootfs-latest.tar.gz
		fi

		# download failed?
		if [ -d "$build_rootfs_dir" ]; then
			build_rootfs_dir="$(realpath "$build_rootfs_dir")"
			info "OK: Linking $build_rootfs_setup_dir to $build_rootfs_dir"
			ln -sf "$build_rootfs_dir" "$build_rootfs_setup_dir" \
				|| soft_error "symbolic link failed: build_rootfs"
		elif [ -n "$build_rootfs_dir" ]; then
			soft_error "could not find $build_rootfs_dir"
		fi
	elif [ -L "$build_rootfs_setup_dir" ]; then
		info "OK: $build_rootfs_setup_dir is a symbolic link to $(readlink "$build_rootfs_setup_dir")"
	else
		info "OK: $build_rootfs_setup_dir exists"
	fi

	local imx_boot_setup_dir="$SECUREBOOT_DIR/imx-boot"
	if [ ! -e "$imx_boot_setup_dir" ]; then
		info "OK: Linking $imx_boot_setup_dir to $IMXBOOT_DIR"
		ln -sf "$IMXBOOT_DIR" "$imx_boot_setup_dir" \
			|| soft_error "symbolic link failed: imx-boot"
	elif [ -L "$imx_boot_setup_dir" ]; then
		info "OK: $imx_boot_setup_dir is a symbolic link to $(readlink "$imx_boot_setup_dir")"
	else
		info "OK: $imx_boot_setup_dir exists"
	fi

	command -v mkswu > /dev/null || soft_error "could not find mkswu command"

	[ -z "$ERRORS" ] || exit 1
	info "Setup $SECUREBOOT_DIR successfully. Run '$0 secureboot_init' next."
}

download_linux_apk() {
	# $apk_dir must be set by parent
	local linux_apk="linux-at-x2-latest.apk"
	local apk_url="https://download.atmark-techno.com/armadillo-iot-g4/baseos/$linux_apk"
	local apk_file="$apk_dir/$linux_apk"

	mkdir -p "$apk_dir" || exit

	info "Downloading $apk_url"
	to_stdout curl -o "$apk_file" "$apk_url" \
		|| error "failed to download $apk_url"
	info

	tar --warning=no-unknown-keyword -xf "$apk_file" -C "$apk_dir" \
		|| error "Could not extract $apk_url"

	rm -f "$apk_file"

	# check package is sane
	[ -d "$apk_dir/boot" ] || error "could not find $apk_dir/boot"
	[ -d "$apk_dir/lib" ] || error "could not find $apk_dir/lib"
}

# This is verbose, so redirect logs early
check_or_download_linux() {
	local apk_dir="$TMPDIR/linux_apk"

	redirect_logs

	# use downloaded linux image/dtb/modules only if unset
	if ! [ -e "$LINUX_IMAGE" ]; then
		echo "Using linux kernel from apk"
		# modules must match image
		LINUX_IMAGE="$apk_dir/boot/Image"
		LINUX_MODULES="$apk_dir/lib/modules"
	fi
	if ! [ -e "$LINUX_DTB" ]; then
		echo "Using linux DTB from apk"
		LINUX_DTB="$apk_dir/boot/armadillo_iotg_g4.dtb"
	fi

	if ! [ -e "$LINUX_IMAGE" ] || ! [ -e "$LINUX_DTB" ]; then
		download_linux_apk
	fi
	# ignore missing modules/initrd
	[ -e "$LINUX_MODULES" ] || LINUX_MODULES=""
	[ -e "$LINUX_INITRD" ] || LINUX_INITRD=""
	case "$LINUX_INITRD_TYPE" in
	""|none|plain|mmc) ;;
	verity)
		[ -n "$LINUX_INITRD_VERITY_CERT" ] \
			|| error "Please set LINUX_INITRD_VERITY_CERT"
		[ -e "$LINUX_INITRD_VERITY_CERT" ] \
			|| error "LINUX_INITRD_VERITY_CERT '$LINUX_INITRD_VERITY_CERT' not found"
		;;
	*) error "invalid LINUX_INITRD_TYPE: $LINUX_INITRD_TYPE (mustbe none/plain/mmc/verity)";;
	esac

	for overlay in "${LINUX_DTB_OVERLAYS[@]}" "${LINUX_DTB_OVERLAYS_PREAPPLY[@]}"; do
		if [[ "${overlay#/}" = "$overlay" ]]; then
			# relative to LINUX_DTB dir if no / in path
			overlay="${LINUX_DTB%/*}/$overlay"
		fi
		[[ -e "$overlay" ]] \
			|| error "$overlay not found in ${LINUX_DTB%/*}"
	done

	echo
	echo "The following files will be used in the subsequent processes."
	echo "image: $LINUX_IMAGE"
	echo "dtb: $LINUX_DTB"
	for dtbo in "${LINUX_DTB_OVERLAYS_PREAPPLY[@]}"; do
		if [[ "${dtbo#/}" != "$dtbo" ]]; then
			echo "dtb_preapplay: $dtbo"
		else
			echo "dtb_preapplay: ${LINUX_DTB%/*}/$dtbo"
		fi
	done
	for dtbo in "${LINUX_DTB_OVERLAYS[@]}"; do
		if [[ "${dtbo#/}" != "$dtbo" ]]; then
			echo "dtb: $dtbo"
		else
			echo "dtb: ${LINUX_DTB%/*}/$dtbo"
		fi
	done
	if [ -n "$LINUX_INITRD" ]; then
		echo "initrd: $LINUX_INITRD"
	elif [ "$LINUX_INITRD_TYPE" = none ]; then
		echo "initrd disabled"
	elif ! [ -e "$SECUREBOOT_DIR/build-rootfs" ]; then
		error "initrd not provided and build-rootfs does not exist, please run $0 setup"
	elif [ -n "$LINUX_INITRD_TYPE" ]; then
		echo "initrd: built for $LINUX_INITRD_TYPE"
	fi
}

download_installer_image() {
	local url=https://armadillo.atmark-techno.com/files/downloads/armadillo-iot-g4/image/baseos-x2-installer-latest.zip
	local work_dir="$TMPDIR/installer.tmp"
	local zip="$work_dir/baseos-x2-installer-latest.zip"

	rm -rf "$work_dir" || exit
	mkdir -p "$work_dir" || exit

	info "Downloading $url"
	to_stdout curl -o "$zip" "$url" \
		|| error "Could not download $url"
	info

	unzip -d "$work_dir" "$zip"

	local img sig
	for img in "$work_dir"/*.img; do :; done
	for sig in "$work_dir"/*.sig; do :; done
	[ -e "$img" ] || error "image file not found in installer archive"

	if ! [ -e "$sig" ]; then
	       warning "signature file not found in installer archive"
	       prompt_yesno n "Continue anyway? [y/N]" \
		       || error "Aborted by user"

	       mv "$img" "$TMPDIR/"
	       rm -rf "$work_dir"
	       installer="$TMPDIR/${img##*/}"
	       return
	fi

	local ca_file="$SECUREBOOT_DIR/build-rootfs/common/image_firstboot/etc/swupdate.pem"
	[ -e "$ca_file" ] \
		|| error "swupdate.pem not found in $SECUREBOOT_DIR/common/image_firstboot/etc, cannot verify installer image -- aborting"

	openssl cms -verify -inform DER -in "$sig" \
			-content "$img" -nosmimecap -binary \
			-CAfile "$ca_file" >/dev/null \
		|| error "Signature verification failed"

	mv "$img" "$TMPDIR/"
	rm -rf "$work_dir"
	installer="$TMPDIR/${img##*/}"
}

find_or_download_installer() {
	installer=$(find "$TMPDIR" -maxdepth 1 -type f -name "baseos-x2-*.img" \
			| sort -Vr | head -n 1)

	if [ -e "$installer" ]; then
		echo "Re-using $installer"
		return
	fi

	download_installer_image
}

secureboot_init() {
	set -o pipefail

	local resources_dir="$SECUREBOOT_DIR/tmp/resources"

	# prepare files for build linux kernel.
	echo "initrd: $LINUX_INITRD"
	echo

	# create srk hash if required
	init_srk
	info

	#write srk_hash to srk_hash.txt
	local srk_hash=""
	srk_hash=$(
		print_srk_fuse \
		| grep "fuse prog -y [6|7] 0" \
		| awk 'BEGIN {FS=" "} {print $6","$7","$8","$9","$10}' \
		| tr -d "\n"
	) || error "could not read srk_hash"
	echo "${srk_hash%,}" > "$TMPDIR/srk_hash.txt"
	echo "created $TMPDIR/srk_hash.txt"
	echo
	# write to "CONFIG_ENV_WRITEABLE_LIST=y" to x2_defconfig.
	# setting "CONFIG_ENV_WRITEABLE_LIST=y" makes U-boot variables
	# without including in CFG_ENV_FLAGS_LIST_STATIC in armadillo_x2.h
	# unchangeble.
	local x2_defconfig=""
	x2_defconfig="$IMXBOOT/uboot-imx/configs/x2_defconfig"
	if ! grep -q "CONFIG_ENV_WRITEABLE_LIST=y" "$x2_defconfig"; then
		echo "CONFIG_ENV_WRITEABLE_LIST=y" >> "$x2_defconfig" \
			|| error "could not set CONFIG_ENV_WRITEABLE_LIST=y in $x2_defconfig"
		echo "set CONFIG_ENV_WRITEABLE_LIST=y in $x2_defconfig"
		echo
	fi
	# write to "CONFIG_BOOTDELAY=-2" to x2_defconfig.
	# setting "CONFIG_BOOTDELAY=-2" does not allow boot prompt to be come in.
	if ! grep -q "CONFIG_BOOTDELAY=-2" "$x2_defconfig"; then
		echo "CONFIG_BOOTDELAY=-2" >> "$x2_defconfig" \
			|| error "could not set CONFIG_BOOTDELAY=-2 in $x2_defconfig"
		echo "set CONFIG_BOOTDELAY=-2 in $x2_defconfig"
		echo
	fi
	local armadillo_x2_h=""
	armadillo_x2_h="$IMXBOOT/uboot-imx/include/configs/armadillo_x2.h"
	if ! grep -q "loadbootscript=false" "$armadillo_x2_h"; then
		sed -i -e 's/loadbootscript=.*\\0/loadbootscript=false\\0/' "$armadillo_x2_h" \
			|| error "could not set loadbootscript=false in $armadillo_x2_h"
		echo "set loadbootscript=false in $armadillo_x2_h"
		echo
	fi

	# create signed boot loader
	imxboot
	[ -e "$OUT/imx-boot_$IMXBOOT_TARGET.signed" ] \
		|| error "$OUT/imx-boot_$IMXBOOT_TARGET.signed is not found"
	info

	# create signed linux kernel image
	rm -f "$OUT/Image.signed"
	# build without initrd for first phase
	LINUX_INITRD="" LINUX_INITRD_TYPE=none \
		linux
	[ -e "$OUT/Image.signed" ] || error "$OUT/Image.signed was not created as expected"
	info

	# init secureboot_x2/secureboot/resources directory.
	rm -rf "$resources_dir"
	mkdir -p "$resources_dir/lib"
	echo "$resources_dir initialized."

	# copy linux kernel image and others to resources directory.
	# resources dir will be included in SWU image.
	echo "copying following files to $resources_dir:"
	echo "resources/lib:"
	cp -r "$LINUX_MODULES" "$resources_dir/lib" \
		|| error "could not copy $LINUX_MODULES to $resources_dir/lib"
	echo " - $LINUX_MODULES"
	echo

	LINUX_OUT="$OUT/Image.signed-mmc"
	LINUX_INITRD_TYPE=mmc
	if [ -e "$LINUX_INITRD" ]; then
		if ! [[ "$LINUX_INITRD" = *.mmc.zst ]]; then
		       info "LINUX_INITRD '$LINUX_INITRD' is not locked to mmc."
		       prompt_yesno n "Continue anyway? [y/N]" \
			       || error "Aborted by user"
		       LINUX_OUT="$OUT/Image.signed"
		fi
	fi

	# build signed linux kernel image with locked initrd.
	rm -f "$LINUX_OUT"
	linux
	[ -e "$LINUX_OUT" ] || error "$LINUX_OUT was not created as expected"

	info

	# create SWU image to write srk hash and install linux kernel image and others
	info "Building 1_write_srk_install_kernel.swu..."
	DTBOS="${LINUX_DTB_OVERLAYS[*]##*/}" \
		mkswu "$IMXBOOT_DIR/secureboot/swu/1_write_srk_install_kernel.desc" \
			-o "$SWU_OUT_DIR/1_write_srk_install_kernel.swu" \
		|| error "could not create $SWU_OUT_DIR/1_write_srk_install_kernel.swu"
	echo

	# create SWU image to close secureboot
	info "Building 2_secureboot_close.swu..."
	mkswu "$IMXBOOT_DIR/secureboot/swu/2_secureboot_close.desc" \
			-o "$SWU_OUT_DIR/2_secureboot_close.swu" \
		|| error "could not create $SWU_OUT_DIR/2_secureboot_close.swu"

	echo

	#build SWU image to encrpt a strage on Armadillo.
	info "Building 3_disk_encryption.swu..."
	mkswu "$IMXBOOT_DIR/secureboot/swu/3_disk_encryption.desc" \
			-o "$SWU_OUT_DIR/3_disk_encryption.swu" \
		|| error "could not create $SWU_OUT_DIR/3_disk_encryption.swu"

	info
	info
	info "Please install SWU images in the following order:"
	info " - $SWU_OUT_DIR/1_write_srk_install_kernel.swu"
	info " - $SWU_OUT_DIR/2_secureboot_close.swu"
	info " - $SWU_OUT_DIR/3_disk_encryption.swu"
}

make_installer() {
	set -o pipefail

	local resources_dir="$TMPDIR/resources_verity"
	local cert_pem

	find_or_download_installer

	# create key.pem and cert.pem for verity check
	echo "creating key.pem and cert.pem"
	cert_pem="$TMPDIR/installer_verity_cert.pem"
	key_pem="$TMPDIR/installer_verity_key.pem"
	cn="${cn:-Secure boot installer $(\
		tr -dc "a-zA-Z0-9" < /dev/urandom| head -c 10)}"
	echo "use cn: $cn"
	echo
	openssl req -x509 -nodes -days 3650 -newkey ec \
		-pkeyopt ec_paramgen_curve:prime256v1 \
		-keyout "$key_pem" \
		-out "$cert_pem" \
		-subj /CN="$cn" \
		|| error "could not create $key_pem and $cert_pem"
	echo "created $cert_pem"
	echo "created $key_pem"
	echo

	# Always rebuild verity initrd (could not possibly guess the keys we just created)
	LINUX_INITRD_TYPE=verity
	LINUX_INITRD_VERITY_CERT="$cert_pem"
	LINUX_INITRD=""

	# sign verity kernel with that initrd
	rm -f "$OUT/Image.signed-verity"
	linux
	[ -e "$OUT/Image.signed-verity" ] || error "Verity image not created as expected"
	info

	# init secureboot_x2/secureboot/resources_verity directory.
	rm -rf "$resources_dir"
	mkdir -p "$resources_dir/sd_copy/boot" || exit
	cp "$OUT/Image.signed-verity" "$resources_dir/sd_copy/boot/Image" \
		|| error "could not copy $LINUX_INITRD to $resources_dir/sd_copy/boot/Image"
	cp "$cert_pem" "$key_pem" "$resources_dir" || exit

	# build SWU image to encrpt a strage on Armadillo.
	info "Building SWU..."
	echo "secureboot_make_installer.swu"
	INSTALLER_IMAGE="$installer" resources_dir="$resources_dir" \
		mkswu "$IMXBOOT_DIR/secureboot/swu/secureboot_make_installer.desc" \
			-o "$SWU_OUT_DIR/secureboot_make_installer.swu" \
		|| error "could not create $SWU_OUT_DIR/secureboot_make_installer.swu"
	info
	info
	info "Install following SWU image with an USB drive plugged in:"
	info " - $SWU_OUT_DIR/secureboot_make_installer.swu"
}

help() {
	echo "Usage: $0 action"
	echo
	echo "Where action is one of:"
	echo "init_srk: initialize super root key"
	echo "print_srk_fuse: Show uboot commands to burn fuses for generated key"
	echo "imxboot: build and sign imx-boot image"
	echo "imxboot_enc: build, encrypt and sign imx-boot image."
	echo "             Note this requires special handling to install."
	echo "linux: build a fit image from configured linux files and sign it."
	echo "       --image also can specify Linux kernel image path."
	echo "       --initrd also can specify locked initrd path."
	echo "       --dtb also can specify dtb path."
	echo "       --dtbo also can specify dtbo path."
	echo "setup: build an environment for secureboot setup."
	echo "secureboot_init: build SWU images to enable secureboot and encrypt"
	echo "                 a storage."
	echo "                 It takes the same arguments as linux, "
	echo "                 besides the listed arguments."
	echo "                 --modules also can specify modules path."
	echo "make_installer: build SWU images to make a installer disk image"
	echo "                for secureboot."
	echo "                It takes the same arguments as linux, besides the"
	echo "                listed arguments."
	echo "                --modules also can specify modules path."
	echo "                --cn also can specify CN for openssl. default is ten"
	echo "                random strings."
	echo
	echo "See $IMXBOOT_DIR/secureboot.conf for configuration settings"
}

main() {
	local mode=""
	local arg=""
	local dtbo_seen=""

	(( $# == 0 )) && help

	while [ "$#" -ge 1 ]; do
		arg="$1"
		case "$arg" in
		imxboot|imxboot_enc|linux|init_srk|print_srk_fuse|setup|secureboot_init|make_installer)
			mode="$arg"
			shift;;
		help|-h|--help) help; exit 0;;
		--image)
			[ "$#" -ge 2 ] || error "$arg requires an argument"
			LINUX_IMAGE="$2"
			shift 2
			;;
		--initrd)
			[ "$#" -ge 2 ] || error "$arg requires an argument"
			LINUX_INITRD="$2"
			shift 2
			;;
		--dtb)
			[ "$#" -ge 2 ] || error "$arg requires an argument"
			LINUX_DTB="$2"
			shift 2
			;;
		--dtbo)
			[ "$#" -ge 2 ] || error "$arg requires an argument"
			if [ -z "$dtbo_seen" ]; then
				dtbo_seen=1
				LINUX_DTB_OVERLAYS_PREAPPLY=()
				LINUX_DTB_OVERLAYS=()
			fi
			LINUX_DTB_OVERLAYS_PREAPPLY+=( "$2" )
			shift 2
			;;
		--modules)
			[ "$#" -ge 2 ] || error "$arg requires an argument"
			LINUX_MODULES="$2"
			shift 2
			;;
		--cn)
			[ "$#" -ge 2 ] || error "$arg requires an argument"
			case "$2" in
			*/*)
				error "--cn could not specify string included in '/' character."
				;;
			esac
			cn="$2"
			shift 2
			;;
		*) warning "Unrecognized option $arg"; echo; help; exit 1;;
		esac
	done

	pre_checks "$mode"

	"$mode"
}

main "$@"
